#ifndef __TEXTURE__
#define __TEXTURE__

#include <GlewOpenGL.h>
#include "Dim2D.h"
#include "TextureData.h"

namespace apig {

    /** This class represents an OpenGL texture (1D, 2D, 3D, cube-map). */
    class Texture {
        public:
            // le constructeur peut etre appele hors d'un contexte GL :
            Texture() : id(0), texData(NULL), loaded(false) {}
            Texture(GLuint id, GLenum texMode) : id(id), texData(NULL), texMode(texMode), loaded(hasContent()) {}
            Texture(const TextureData *texData, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);
            Texture(GLenum texMode, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);     // alloue un id de texture
            
            // chargement de la texture en memoire (doit etre appele dans un contexte GL) :
            void load();                        // charge <texData> en memoire GPU en utilisant le format par defaut de <texData>
            void load(GLint internalFormat);    // de mme mais avec le format precise
            bool isLoaded() const { return loaded; }
            
            // les methodes suivantes sont a appeler une fois que la texture est chargee :
            void activate() const;
            void deactivate() const;
            void destroy();
            
            // pour l'utilisation de plusieurs d'unites de textures :
            void activate(int texNum) const;    // associe la texture a l'unite de texture 'texNum'
            void deactivate(int texNum) const;  // desactive l'unite de texture 'texNum'
            
            // reglage des parametres de la texture (un contexte GL doit etre actif) :
            // la texture doit avoir ete creee (methode load())
            void setInterpMode(GLenum interpMode);
            void setWrapMode(GLenum wrapMode);
            void setBorderColor(const float *color) const;
            void setBorderColor(float r, float g, float b, float a = 1) const;
            
            GLuint getID()   const { return id; }
            GLenum getMode() const { return texMode; }
            
            // methodes de creation de textures vides :
            // <internalFormat> : tables 3.15 et 3.16 de la spec OpenGL 2.0
            static Texture createTex1D(int w,               GLint internalFormat = GL_RGBA, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);
            static Texture createTex2D(int w, int h,        GLint internalFormat = GL_RGBA, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);
            static Texture createTex3D(int w, int h, int d, GLint internalFormat = GL_RGBA, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);
            static Texture createTexCube(int n,             GLint internalFormat = GL_RGBA, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);
    
            void copyReadBuffer(GLenum internalFormat, int w, int h, int x=0, int y=0);    // copie le contenu du render-buffer courant dans la texture
            
            // methodes utiles :
            void bind() const;          // attache cette texture comme texture courante (associee a TEX_MODE)
            int getParam(GLenum paramName) const;
            int getLevelParam(GLenum paramName, int level = 0) const { return getLevelParam(texMode, paramName, level); }
            int getLevelParam(int target, GLenum paramName, int level = 0) const;
            int getWidth() const  { return getLevelParam(GL_TEXTURE_WIDTH);  }
            int getHeight() const { return getLevelParam(GL_TEXTURE_HEIGHT); }
            int getDepth() const  { return getLevelParam(GL_TEXTURE_DEPTH);  }
            // effets de bord : texture bindee
            
            // les methodes suivantes calculent la taille 2D de l'image qui peut etre rendue dans cette texture avec un frame-buffer :
            Dim2D renderSize(int level = 0) const { return renderSize(texMode, level); }
            Dim2D renderSize(GLenum target, int level = 0) const;
            
        private:
            void generateID();              // alloue un nouvel identifiant. precondition : id <= 0
            void init() const;              // initialise les parametres de la texture (effet de bord : la texture est bindee)
            void updateInterpMode() const;
            void updateWrapMode() const;
            bool hasContent() const;    // effet de bord : texture bindee
            
        private:
            GLuint id;
            const TextureData *texData;
            mutable GLenum texMode,     // GL_TEXTURE_1D, GL_TEXTURE_2D, GL_TEXTURE_3D, GL_TEXTURE_CUBE_MAP
                   interpMode,  // GL_NEAREST, GL_LINEAR
                   wrapMode;    // GL_REPEAT, GL_CLAMP, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_BORDER, GL_MIRRORED_REPEAT
            bool loaded;    // true ssi le contenu de la texture est defini
        };
    
    }

#endif

